package http

import "net"
import "log"
import "fmt"
import "time"
import "path"
import "io/fs"
import "errors"
import "strings"
import "strconv"
import "context"
import "net/http"
import "path/filepath"
import "git.tebibyte.media/sashakoshka/step"
import "git.tebibyte.media/sashakoshka/goutil/sync"
import "git.tebibyte.media/sashakoshka/goutil/container"

var errRateLimit = errors.New("rate limit")

type rateLimitEntry struct {
	time time.Time
	duration time.Duration
}

type DirectoryData struct {
	Name    string
	Entries []fs.DirEntry
}

type Handler struct {
	Environment         *step.Environment
	Directories         bool
	StepExt             ucontainer.Set[string]
	Index               []string
	ErrorDocument       string
	DirectoryDocument   string
	DenyAll             ucontainer.Set[string]
	RateLimit           time.Duration
	TrustXForwardedFor  bool
	TrustCFConnectingIP bool

	rateLimitMap usync.Locker[map[string] rateLimitEntry]
}

func (this *Handler) Init (ctx context.Context) error {
	this.rateLimitMap.Set(make(map[string] rateLimitEntry))
	return nil
}

func (this *Handler) ServeHTTP (res http.ResponseWriter, req *http.Request) {
	remoteAddr    := req.RemoteAddr
	remoteAddrStr := req.RemoteAddr
	if addr := req.Header.Get("CF-Connecting-IP"); addr != "" {
		remoteAddr = addr
		if this.TrustCFConnectingIP {
			remoteAddrStr = fmt.Sprintf("%s --CF-> %s", addr, req.RemoteAddr)
		}
	} else if addr := req.Header.Get("X-Forwarded-For"); addr != "" {
		remoteAddr = addr
		if this.TrustXForwardedFor {
			remoteAddrStr = fmt.Sprintf("%s --??-> %s", addr, req.RemoteAddr)
		}
	}
	log.Println("(i)", req.Method, req.URL, "from", remoteAddrStr)
	filesystem := this.Environment.GetFS()

	// rate limit
	err := this.rateLimit(req, remoteAddr, "", this.RateLimit)
	if err != nil {
		this.serveError(res, req, http.StatusTooManyRequests, req.URL, false)
		return
	}

	// normalize path
	pat := req.URL.Path
	if !strings.HasPrefix(pat, "/") {
		pat = "/" + pat
		req.URL.Path = pat
	}
	hasTrailingSlash := strings.HasSuffix(pat, "/")
	pat = path.Clean(req.URL.Path)
	name := pathToName(pat)

	// access control
	if this.DenyAll.Has(name) {
		this.serveError(res, req, http.StatusForbidden, req.URL, false)
		return
	}

	info, err := statFile(filesystem, name)
	if err != nil {
		this.serveError(res, req, http.StatusNotFound, req.URL, false)
		return
	}
	if info.IsDir() {
		// ensure the path ends with a /
		if !hasTrailingSlash {
			http.Redirect(res, req, pat + "/", http.StatusMovedPermanently)
			return
		}
	
		// try to find an index
		for _, base := range this.Index {
			currentPath := path.Join(pat, base)
			info, err := statFile(filesystem, pathToName(currentPath))
			if err != nil   { continue }
			if info.IsDir() { continue }
			this.serveFile(res, req, currentPath, remoteAddr)
			return
		}

		if !this.Directories {
			this.serveError(res, req, http.StatusForbidden, req.URL, false)
			return
		}

		this.serveDirectory(res, req, pat, remoteAddr)
		return
	}

	this.serveFile(res, req, pat, remoteAddr)
}

func (this *Handler) Trim () {
	table, done := this.rateLimitMap.Borrow()
	defer done()
	if len(table) < 128 { return }
	keys := make([]string, 0, len(table) / 3)
	for key, entry := range table {
		duration := time.Hour
		if entry.duration != 0 {
			duration = entry.duration
		}
		if time.Since(entry.time) > duration {
			keys = append(keys, key)
		}
		
	}
	for _, key := range keys {
		delete(table, key)
	}
}

func (this *Handler) serveFile (
	res http.ResponseWriter,
	req *http.Request,
	pat string,
	remoteAddr string,
) {
	name := pathToName(pat)
	if !this.StepExt.Has(filepath.Ext(name)) {
		// just a normal file
		http.ServeFileFS(res, req, this.Environment.GetFS(), name)
		return
	}
	this.serveDocument(res, req, name, remoteAddr)	
}

func (this *Handler) serveDirectory (
	res http.ResponseWriter,
	req *http.Request,
	pat string,
	remoteAddr string,
) {
	name := pathToName(pat)
	entries, err := readDir(this.Environment.GetFS(), name)
	if err != nil {
		this.serveError(res, req, http.StatusInternalServerError, err, false)
		return
	}
	if this.DirectoryDocument == "" {
		this.serveFile(res, req, pat, remoteAddr)
		return
	}
	document, err := this.Environment.Load(this.DirectoryDocument)
	if err != nil {
		this.serveError(res, req, http.StatusInternalServerError, err, false)
		return
	}
	err = document.Execute(res, step.ExecutionData {
		Data: DirectoryData {
			Name:    name,
			Entries: entries,
		},
	})
	if err != nil {
		this.serveError(res, req, http.StatusInternalServerError, err, false)
		return
	}
}

func (this *Handler) serveDocument (
	res http.ResponseWriter,
	req *http.Request,
	name string,
	remoteAddr string,
) {
	// parse
	document, err := this.Environment.Load(name)
	if err != nil {
		this.serveError(res, req, http.StatusInternalServerError, err, false)
		return
	}

	// rate limit (again)
	if rateLimit, err := strconv.ParseFloat(document.Meta.Get("rate-limit"), 64); err == nil {
		err := this.rateLimit(req, remoteAddr, name, time.Duration(rateLimit * float64(time.Second)))
		if err != nil {
			this.serveError(res, req, http.StatusTooManyRequests, req.URL, false)
			return
		}
	}
	
	// set up HTTP response recorder
	recorder := HTTPResponseRecorder { }
	resetRecorder := func () {
		recorder.Reset()
		recorder.Head = res.Header().Clone()
	}
	if contentType := document.Meta.Get("content-type"); contentType != "" {
		recorder.Header().Set("Content-Type", contentType)
	}
	if status := document.Meta.Get("status"); status != "" {
		if status, err := strconv.Atoi(status); err == nil {
			recorder.Status = status
		}
	}

	// execute document
	data := HTTPData { }
	data.Res = WrappedResponseWriter {
		responseWriter: &recorder,
		resetFunc: resetRecorder,
		Header: WrappedHeader {
			Header: recorder.Header(),
		},
	}
	data.Req = req
	err = document.Execute(&recorder, step.ExecutionData {
		Data: data,
	})
	if errors.Is(err, step.ErrExecutionCanceled) { err = nil }
	var httpError Error
	if errors.As(err, &httpError) {
		this.serveError (
			res, req,
			httpError.Status, httpError.Message, false)
		return
	}
	var httpRedirect Redirect
	if errors.As(err, &httpRedirect) {
		http.Redirect(res, req, httpRedirect.Location, httpRedirect.Status)
		return
	}
	if err != nil {
		this.serveError (
			res, req,
			http.StatusInternalServerError, err, false)
		return
	}

	// play back recorded response
	err = recorder.Play(res)
	if err != nil {
		this.logErr(name, err)
	}
}

func (this *Handler) serveError (
	res http.ResponseWriter,
	req *http.Request,
	status int,
	message any,
	safeMode bool,
) {
	log.Printf("ERR %d %s: %v\n", status, http.StatusText(status), message)
	if safeMode || this.ErrorDocument == "" {
		res.Header().Add("Content-Type", "text/plain")
		res.WriteHeader(status)
		// TODO: allow customization with templates
		if message == nil {
			fmt.Fprintf(res, "%d %s\n", status, http.StatusText(status))
		} else {
			fmt.Fprintf(res, "%d %s: %v\n", status, http.StatusText(status), message)
		}
		return
	}

	document, err := this.Environment.Load(this.ErrorDocument)
	if err != nil {
		this.serveError(res, req, http.StatusInternalServerError, err, true)
		return
	}
	err = document.Execute(res, step.ExecutionData {
		Data: Error {
			Status:  status,
			Message: message,
		},
	})
	if err != nil {
		this.serveError(res, req, http.StatusInternalServerError, err, true)
		return
	}
}

func (this *Handler) logErr (name string, err error) {
	log.Printf("ERR %s: %v\n", name, err)
}

func (this *Handler) rateLimit (req *http.Request, remoteAddr string, resource string, duration time.Duration) error {
	host, _, _ := net.SplitHostPort(remoteAddr)
	key := fmt.Sprintf("%s|%s", host, resource)
	table, done := this.rateLimitMap.Borrow()
	defer done()
	if entry, ok := table[key]; ok {
		if time.Since(entry.time) < duration {
			return errRateLimit
		}
	}
	table[key] = rateLimitEntry {
		time: time.Now(),
		duration: duration,
	}
	return nil
}

func statFile (filesystem fs.FS, name string) (fs.FileInfo, error) {
	if filesystem, ok := filesystem.(fs.StatFS); ok {
		return filesystem.Stat(name)
	}
	file, err := filesystem.Open(name)
	if err != nil { return nil, err }
	defer file.Close()
	return file.Stat()
}

func pathToName (pat string) string {
	if strings.HasPrefix(pat, "/") {
		pat = strings.TrimPrefix(pat, "/")
	}
	if pat == "" {
		return "."
	}
	return pat
}

func readDir (filesystem fs.FS, name string) ([]fs.DirEntry, error) {
	if filesystem, ok := filesystem.(fs.ReadDirFS); ok {
		return filesystem.ReadDir(name)
	}
	dir, err := filesystem.Open(name)
	if err != nil { return nil, err }
	defer dir.Close()
	if dir, ok := dir.(fs.ReadDirFile); ok {
		return dir.ReadDir(0)
	}
	return nil, nil
}
